@@ -106,7 +106,7 @@ using bst_uint = uint32_t;  // NOLINT
 /*! \brief integer type. */
 using bst_int = int32_t;    // NOLINT
 /*! \brief unsigned long integers */
-using bst_ulong = uint64_t;
+using bst_ulong = uint64_t;  // NOLINT
 /*! \brief float type, used for storing statistics */
 using bst_float = float;  // NOLINT
 
@@ -135,15 +135,15 @@ class GradientPairInternal {
   /*! \brief second order gradient statistics */
   T hess_;
 
-  XGBOOST_DEVICE void SetGrad(float g) { grad_ = g; }
-  XGBOOST_DEVICE void SetHess(float h) { hess_ = h; }
+  XGBOOST_DEVICE void SetGrad(T g) { grad_ = g; }
+  XGBOOST_DEVICE void SetHess(T h) { hess_ = h; }
 
  public:
   using ValueT = T;
 
   XGBOOST_DEVICE GradientPairInternal() : grad_(0), hess_(0) {}
 
-  XGBOOST_DEVICE GradientPairInternal(float grad, float hess) {
+  XGBOOST_DEVICE GradientPairInternal(T grad, T hess) {
     SetGrad(grad);
     SetHess(hess);
   }
@@ -160,8 +160,8 @@ class GradientPairInternal {
     SetHess(g.GetHess());
   }
 
-  XGBOOST_DEVICE float GetGrad() const { return grad_; }
-  XGBOOST_DEVICE float GetHess() const { return hess_; }
+  XGBOOST_DEVICE T GetGrad() const { return grad_; }
+  XGBOOST_DEVICE T GetHess() const { return hess_; }
 
   XGBOOST_DEVICE GradientPairInternal<T> &operator+=(
       const GradientPairInternal<T> &rhs) {
@@ -193,6 +193,36 @@ class GradientPairInternal {
     return g;
   }
 
+  XGBOOST_DEVICE GradientPairInternal<T> &operator*=(float multiplier) {
+    grad_ *= multiplier;
+    hess_ *= multiplier;
+    return *this;
+  }
+
+  XGBOOST_DEVICE GradientPairInternal<T> operator*(float multiplier) const {
+    GradientPairInternal<T> g;
+    g.grad_ = grad_ * multiplier;
+    g.hess_ = hess_ * multiplier;
+    return g;
+  }
+
+  XGBOOST_DEVICE GradientPairInternal<T> &operator/=(float divisor) {
+    grad_ /= divisor;
+    hess_ /= divisor;
+    return *this;
+  }
+
+  XGBOOST_DEVICE GradientPairInternal<T> operator/(float divisor) const {
+    GradientPairInternal<T> g;
+    g.grad_ = grad_ / divisor;
+    g.hess_ = hess_ / divisor;
+    return g;
+  }
+
+  XGBOOST_DEVICE bool operator==(const GradientPairInternal<T> &rhs) const {
+    return grad_ == rhs.grad_ && hess_ == rhs.hess_;
+  }
+
   XGBOOST_DEVICE explicit GradientPairInternal(int value) {
     *this = GradientPairInternal<T>(static_cast<float>(value),
                                   static_cast<float>(value));
@@ -204,24 +234,6 @@ class GradientPairInternal {
     return os;
   }
 };
-
-template<>
-inline XGBOOST_DEVICE float GradientPairInternal<int64_t>::GetGrad() const {
-  return grad_ * 1e-4f;
-}
-template<>
-inline XGBOOST_DEVICE float GradientPairInternal<int64_t>::GetHess() const {
-  return hess_ * 1e-4f;
-}
-template<>
-inline XGBOOST_DEVICE void GradientPairInternal<int64_t>::SetGrad(float g) {
-  grad_ = static_cast<int64_t>(std::round(g * 1e4));
-}
-template<>
-inline XGBOOST_DEVICE void GradientPairInternal<int64_t>::SetHess(float h) {
-  hess_ = static_cast<int64_t>(std::round(h * 1e4));
-}
-
 }  // namespace detail
 
 /*! \brief gradient statistics pair usually needed in gradient boosting */
@@ -230,11 +242,6 @@ using GradientPair = detail::GradientPairInternal<float>;
 /*! \brief High precision gradient statistics pair */
 using GradientPairPrecise = detail::GradientPairInternal<double>;
 
-/*! \brief High precision gradient statistics pair with integer backed
- * storage. Operators are associative where floating point versions are not
- * associative. */
-using GradientPairInteger = detail::GradientPairInternal<int64_t>;
-
 using Args = std::vector<std::pair<std::string, std::string> >;
 
 /*! \brief small eps gap for minimum split decision. */
