@@ -1,5 +1,5 @@
 /*!
- * Copyright 2017-2019 by Contributors
+ * Copyright 2017-2020 by Contributors
  * \file hist_util.cc
  */
 #include <dmlc/timer.h>
@@ -11,10 +11,10 @@
 
 #include "xgboost/base.h"
 #include "../common/common.h"
-#include "./hist_util.h"
-#include "./random.h"
-#include "./column_matrix.h"
-#include "./quantile.h"
+#include "hist_util.h"
+#include "random.h"
+#include "column_matrix.h"
+#include "quantile.h"
 #include "./../tree/updater_quantile_hist.h"
 
 #if defined(XGBOOST_MM_PREFETCH_PRESENT)
@@ -29,9 +29,92 @@
 namespace xgboost {
 namespace common {
 
+template<typename BinIdxType>
+void GHistIndexMatrix::SetIndexDataForDense(common::Span<BinIdxType> index_data_span,
+                                    size_t batch_threads, const SparsePage& batch,
+                                    size_t rbegin, common::Span<const uint32_t> offsets_span,
+                                    size_t nbins) {
+  const xgboost::Entry* data_ptr = batch.data.HostVector().data();
+  const std::vector<bst_row_t>& offset_vec = batch.offset.HostVector();
+  const size_t batch_size = batch.Size();
+  CHECK_LT(batch_size, offset_vec.size());
+  BinIdxType* index_data = index_data_span.data();
+  const uint32_t* offsets = offsets_span.data();
+  #pragma omp parallel for num_threads(batch_threads) schedule(static)
+    for (omp_ulong i = 0; i < batch_size; ++i) {
+      const int tid = omp_get_thread_num();
+      size_t ibegin = row_ptr[rbegin + i];
+      size_t iend = row_ptr[rbegin + i + 1];
+      const size_t size = offset_vec[i + 1] - offset_vec[i];
+      SparsePage::Inst inst = {data_ptr + offset_vec[i], size};
+      CHECK_EQ(ibegin + inst.size(), iend);
+      for (bst_uint j = 0; j < inst.size(); ++j) {
+        uint32_t idx = cut.SearchBin(inst[j]);
+        index_data[ibegin + j] = static_cast<BinIdxType>(idx - offsets[j]);
+        ++hit_count_tloc_[tid * nbins + idx];
+      }
+    }
+}
+template void GHistIndexMatrix::SetIndexDataForDense(common::Span<uint8_t> index_data_span,
+                                             size_t batch_threads, const SparsePage& batch,
+                                             size_t rbegin,
+                                             common::Span<const uint32_t> offsets_span,
+                                             size_t nbins);
+template void GHistIndexMatrix::SetIndexDataForDense(common::Span<uint16_t> index_data_span,
+                                             size_t batch_threads, const SparsePage& batch,
+                                             size_t rbegin,
+                                             common::Span<const uint32_t> offsets_span,
+                                             size_t nbins);
+template void GHistIndexMatrix::SetIndexDataForDense(common::Span<uint32_t> index_data_span,
+                                             size_t batch_threads, const SparsePage& batch,
+                                             size_t rbegin,
+                                             common::Span<const uint32_t> offsets_span,
+                                             size_t nbins);
+
+void GHistIndexMatrix::SetIndexDataForSparse(common::Span<uint32_t> index_data_span,
+                                                 size_t batch_threads,
+                                                 const SparsePage& batch, size_t rbegin,
+                                                 size_t nbins) {
+  const xgboost::Entry* data_ptr = batch.data.HostVector().data();
+  const std::vector<bst_row_t>& offset_vec = batch.offset.HostVector();
+  const size_t batch_size = batch.Size();
+  CHECK_LT(batch_size, offset_vec.size());
+  uint32_t* index_data = index_data_span.data();
+  #pragma omp parallel for num_threads(batch_threads) schedule(static)
+    for (omp_ulong i = 0; i < batch_size; ++i) {
+      const int tid = omp_get_thread_num();
+      size_t ibegin = row_ptr[rbegin + i];
+      size_t iend = row_ptr[rbegin + i + 1];
+      const size_t size = offset_vec[i + 1] - offset_vec[i];
+      SparsePage::Inst inst = {data_ptr + offset_vec[i], size};
+      CHECK_EQ(ibegin + inst.size(), iend);
+      for (bst_uint j = 0; j < inst.size(); ++j) {
+        uint32_t idx = cut.SearchBin(inst[j]);
+        index_data[ibegin + j] = idx;
+        ++hit_count_tloc_[tid * nbins + idx];
+      }
+    }
+}
+
+void GHistIndexMatrix::ResizeIndex(const size_t rbegin, const SparsePage& batch,
+                                   const size_t n_offsets, const size_t n_index,
+                                   const bool isDense) {
+  if ((max_num_bins - 1 <= static_cast<int>(std::numeric_limits<uint8_t>::max())) && isDense) {
+    index.SetBinTypeSize(kUint8BinsTypeSize);
+    index.Resize((sizeof(uint8_t)) * n_index);
+  } else if ((max_num_bins - 1 > static_cast<int>(std::numeric_limits<uint8_t>::max())  &&
+    max_num_bins - 1 <= static_cast<int>(std::numeric_limits<uint16_t>::max())) && isDense) {
+    index.SetBinTypeSize(kUint16BinsTypeSize);
+    index.Resize((sizeof(uint16_t)) * n_index);
+  } else {
+    index.SetBinTypeSize(kUint32BinsTypeSize);
+    index.Resize((sizeof(uint32_t)) * n_index);
+  }
+}
+
 HistogramCuts::HistogramCuts() {
   monitor_.Init(__FUNCTION__);
-  cut_ptrs_.emplace_back(0);
+  cut_ptrs_.HostVector().emplace_back(0);
 }
 
 // Dispatch to specific builder.
@@ -52,7 +135,7 @@ void HistogramCuts::Build(DMatrix* dmat, uint32_t const max_num_bins) {
     DenseCuts cuts(this);
     cuts.Build(dmat, max_num_bins);
   }
-  LOG(INFO) << "Total number of hist bins: " << cut_ptrs_.back();
+  LOG(INFO) << "Total number of hist bins: " << cut_ptrs_.HostVector().back();
 }
 
 bool CutsBuilder::UseGroup(DMatrix* dmat) {
@@ -70,23 +153,25 @@ void SparseCuts::SingleThreadBuild(SparsePage const& page, MetaInfo const& info,
                                    bool const use_group_ind,
                                    uint32_t beg_col, uint32_t end_col,
                                    uint32_t thread_id) {
-  using WXQSketch = common::WXQuantileSketch<bst_float, bst_float>;
   CHECK_GE(end_col, beg_col);
   constexpr float kFactor = 8;
 
   // Data groups, used in ranking.
   std::vector<bst_uint> const& group_ptr = info.group_ptr_;
-  p_cuts_->min_vals_.resize(end_col - beg_col, 0);
+  auto &local_min_vals = p_cuts_->min_vals_.HostVector();
+  auto &local_cuts = p_cuts_->cut_values_.HostVector();
+  auto &local_ptrs = p_cuts_->cut_ptrs_.HostVector();
+  local_min_vals.resize(end_col - beg_col, 0);
 
   for (uint32_t col_id = beg_col; col_id < page.Size() && col_id < end_col; ++col_id) {
     // Using a local variable makes things easier, but at the cost of memory trashing.
-    WXQSketch sketch;
+    WQSketch sketch;
     common::Span<xgboost::Entry const> const column = page[col_id];
     uint32_t const n_bins = std::min(static_cast<uint32_t>(column.size()),
                                      max_num_bins);
     if (n_bins == 0) {
       // cut_ptrs_ is initialized with a zero, so there's always an element at the back
-      p_cuts_->cut_ptrs_.emplace_back(p_cuts_->cut_ptrs_.back());
+      local_ptrs.emplace_back(local_ptrs.back());
       continue;
     }
 
@@ -104,26 +189,26 @@ void SparseCuts::SingleThreadBuild(SparsePage const& page, MetaInfo const& info,
       sketch.Push(entry.fvalue, info.GetWeight(weight_ind));
     }
 
-    WXQSketch::SummaryContainer out_summary;
+    WQSketch::SummaryContainer out_summary;
     sketch.GetSummary(&out_summary);
-    WXQSketch::SummaryContainer summary;
-    summary.Reserve(n_bins);
-    summary.SetPrune(out_summary, n_bins);
+    WQSketch::SummaryContainer summary;
+    summary.Reserve(n_bins + 1);
+    summary.SetPrune(out_summary, n_bins + 1);
 
     // Can be use data[1] as the min values so that we don't need to
     // store another array?
     float mval = summary.data[0].value;
-    p_cuts_->min_vals_[col_id - beg_col]  = mval - (fabs(mval) + 1e-5);
+    local_min_vals[col_id - beg_col]  = mval - (fabs(mval) + 1e-5);
 
-    this->AddCutPoint(summary);
+    this->AddCutPoint(summary, max_num_bins);
 
     bst_float cpt = (summary.size > 0) ?
                     summary.data[summary.size - 1].value :
-                    p_cuts_->min_vals_[col_id - beg_col];
+                    local_min_vals[col_id - beg_col];
     cpt += fabs(cpt) + 1e-5;
-    p_cuts_->cut_values_.emplace_back(cpt);
+    local_cuts.emplace_back(cpt);
 
-    p_cuts_->cut_ptrs_.emplace_back(p_cuts_->cut_values_.size());
+    local_ptrs.emplace_back(local_cuts.size());
   }
 }
 
@@ -197,33 +282,40 @@ void SparseCuts::Concat(
     std::vector<std::unique_ptr<SparseCuts>> const& cuts, uint32_t n_cols) {
   monitor_.Start(__FUNCTION__);
   uint32_t nthreads = omp_get_max_threads();
-  p_cuts_->min_vals_.resize(n_cols, std::numeric_limits<float>::max());
+  auto &local_min_vals = p_cuts_->min_vals_.HostVector();
+  auto &local_cuts = p_cuts_->cut_values_.HostVector();
+  auto &local_ptrs = p_cuts_->cut_ptrs_.HostVector();
+  local_min_vals.resize(n_cols, std::numeric_limits<float>::max());
   size_t min_vals_tail = 0;
 
   for (uint32_t t = 0; t < nthreads; ++t) {
+    auto& thread_min_vals = cuts[t]->p_cuts_->min_vals_.HostVector();
+    auto& thread_cuts = cuts[t]->p_cuts_->cut_values_.HostVector();
+    auto& thread_ptrs = cuts[t]->p_cuts_->cut_ptrs_.HostVector();
+
     // concat csc pointers.
-    size_t const old_ptr_size = p_cuts_->cut_ptrs_.size();
-    p_cuts_->cut_ptrs_.resize(
-        cuts[t]->p_cuts_->cut_ptrs_.size() + p_cuts_->cut_ptrs_.size() - 1);
-    size_t const new_icp_size = p_cuts_->cut_ptrs_.size();
-    auto tail = p_cuts_->cut_ptrs_[old_ptr_size-1];
+    size_t const old_ptr_size = local_ptrs.size();
+    local_ptrs.resize(
+        thread_ptrs.size() + local_ptrs.size() - 1);
+    size_t const new_icp_size = local_ptrs.size();
+    auto tail = local_ptrs[old_ptr_size-1];
     for (size_t j = old_ptr_size; j < new_icp_size; ++j) {
-      p_cuts_->cut_ptrs_[j] = tail + cuts[t]->p_cuts_->cut_ptrs_[j-old_ptr_size+1];
+      local_ptrs[j] = tail + thread_ptrs[j-old_ptr_size+1];
     }
     // concat csc values
-    size_t const old_iv_size = p_cuts_->cut_values_.size();
-    p_cuts_->cut_values_.resize(
-        cuts[t]->p_cuts_->cut_values_.size() + p_cuts_->cut_values_.size());
-    size_t const new_iv_size = p_cuts_->cut_values_.size();
+    size_t const old_iv_size = local_cuts.size();
+    local_cuts.resize(
+        thread_cuts.size() + local_cuts.size());
+    size_t const new_iv_size = local_cuts.size();
     for (size_t j = old_iv_size; j < new_iv_size; ++j) {
-      p_cuts_->cut_values_[j] = cuts[t]->p_cuts_->cut_values_[j-old_iv_size];
+      local_cuts[j] = thread_cuts[j-old_iv_size];
     }
     // merge min values
-    for (size_t j = 0; j < cuts[t]->p_cuts_->min_vals_.size(); ++j) {
-      p_cuts_->min_vals_.at(min_vals_tail + j) =
-          std::min(p_cuts_->min_vals_.at(min_vals_tail + j), cuts.at(t)->p_cuts_->min_vals_.at(j));
+    for (size_t j = 0; j < thread_min_vals.size(); ++j) {
+       local_min_vals.at(min_vals_tail + j) =
+          std::min(local_min_vals.at(min_vals_tail + j), thread_min_vals.at(j));
     }
-    min_vals_tail += cuts[t]->p_cuts_->min_vals_.size();
+    min_vals_tail += thread_min_vals.size();
   }
   monitor_.Stop(__FUNCTION__);
 }
@@ -234,7 +326,7 @@ void DenseCuts::Build(DMatrix* p_fmat, uint32_t max_num_bins) {
 
   // safe factor for better accuracy
   constexpr int kFactor = 8;
-  std::vector<WXQSketch> sketchs;
+  std::vector<WQSketch> sketchs;
 
   const int nthread = omp_get_max_threads();
 
@@ -251,7 +343,7 @@ void DenseCuts::Build(DMatrix* p_fmat, uint32_t max_num_bins) {
   size_t const num_groups = group_ptr.size() == 0 ? 0 : group_ptr.size() - 1;
   // Use group index for weights?
   bool const use_group = UseGroup(p_fmat);
-
+  const bool isDense = p_fmat->IsDense();
   for (const auto &batch : p_fmat->GetBatches<SparsePage>()) {
     size_t group_ind = 0;
     if (use_group) {
@@ -276,10 +368,18 @@ void DenseCuts::Build(DMatrix* p_fmat, uint32_t max_num_bins) {
             // move to next group
             group_ind++;
           }
-          for (auto const& entry : inst) {
-            if (entry.index >= begin && entry.index < end) {
-              size_t w_idx = use_group ? group_ind : ridx;
-              sketchs[entry.index].Push(entry.fvalue, info.GetWeight(w_idx));
+          size_t w_idx = use_group ? group_ind : ridx;
+          auto w = info.GetWeight(w_idx);
+          if (isDense) {
+            auto data = inst.data();
+            for (size_t ii = begin; ii < end; ii++) {
+              sketchs[ii].Push(data[ii].fvalue, w);
+            }
+          } else {
+            for (auto const& entry : inst) {
+              if (entry.index >= begin && entry.index < end) {
+                sketchs[entry.index].Push(entry.fvalue, w);
+              }
             }
           }
         }
@@ -287,63 +387,77 @@ void DenseCuts::Build(DMatrix* p_fmat, uint32_t max_num_bins) {
     }
   }
 
-  Init(&sketchs, max_num_bins);
+  Init(&sketchs, max_num_bins, info.num_row_);
   monitor_.Stop(__FUNCTION__);
 }
 
+/**
+ * \param [in,out]  in_sketchs
+ * \param           max_num_bins  The maximum number bins.
+ * \param           max_rows      Number of rows in this DMatrix.
+ */
 void DenseCuts::Init
-(std::vector<WXQSketch>* in_sketchs, uint32_t max_num_bins) {
+(std::vector<WQSketch>* in_sketchs, uint32_t max_num_bins, size_t max_rows) {
   monitor_.Start(__func__);
-  std::vector<WXQSketch>& sketchs = *in_sketchs;
+  std::vector<WQSketch>& sketchs = *in_sketchs;
+
+  // Compute how many cuts samples we need at each node
+  // Do not require more than the number of total rows  in training data
+  // This allows efficient training on wide data
+  size_t global_max_rows = max_rows;
+  rabit::Allreduce<rabit::op::Sum>(&global_max_rows, 1);
   constexpr int kFactor = 8;
+  size_t intermediate_num_cuts =
+      std::min(global_max_rows, static_cast<size_t>(max_num_bins * kFactor));
   // gather the histogram data
-  rabit::SerializeReducer<WXQSketch::SummaryContainer> sreducer;
-  std::vector<WXQSketch::SummaryContainer> summary_array;
+  rabit::SerializeReducer<WQSketch::SummaryContainer> sreducer;
+  std::vector<WQSketch::SummaryContainer> summary_array;
   summary_array.resize(sketchs.size());
   for (size_t i = 0; i < sketchs.size(); ++i) {
-    WXQSketch::SummaryContainer out;
+    WQSketch::SummaryContainer out;
     sketchs[i].GetSummary(&out);
-    summary_array[i].Reserve(max_num_bins * kFactor);
-    summary_array[i].SetPrune(out, max_num_bins * kFactor);
+    summary_array[i].Reserve(intermediate_num_cuts);
+    summary_array[i].SetPrune(out, intermediate_num_cuts);
   }
   CHECK_EQ(summary_array.size(), in_sketchs->size());
-  size_t nbytes = WXQSketch::SummaryContainer::CalcMemCost(max_num_bins * kFactor);
+  size_t nbytes = WQSketch::SummaryContainer::CalcMemCost(intermediate_num_cuts);
   // TODO(chenqin): rabit failure recovery assumes no boostrap onetime call after loadcheckpoint
   // we need to move this allreduce before loadcheckpoint call in future
   sreducer.Allreduce(dmlc::BeginPtr(summary_array), nbytes, summary_array.size());
-  p_cuts_->min_vals_.resize(sketchs.size());
+  p_cuts_->min_vals_.HostVector().resize(sketchs.size());
 
   for (size_t fid = 0; fid < summary_array.size(); ++fid) {
-    WXQSketch::SummaryContainer a;
-    a.Reserve(max_num_bins);
-    a.SetPrune(summary_array[fid], max_num_bins);
+    WQSketch::SummaryContainer a;
+    a.Reserve(max_num_bins + 1);
+    a.SetPrune(summary_array[fid], max_num_bins + 1);
     const bst_float mval = a.data[0].value;
-    p_cuts_->min_vals_[fid] = mval - (fabs(mval) + 1e-5);
-    AddCutPoint(a);
+    p_cuts_->min_vals_.HostVector()[fid] = mval - (fabs(mval) + 1e-5);
+    AddCutPoint(a, max_num_bins);
     // push a value that is greater than anything
     const bst_float cpt
-      = (a.size > 0) ? a.data[a.size - 1].value : p_cuts_->min_vals_[fid];
+      = (a.size > 0) ? a.data[a.size - 1].value : p_cuts_->min_vals_.HostVector()[fid];
     // this must be bigger than last value in a scale
     const bst_float last = cpt + (fabs(cpt) + 1e-5);
-    p_cuts_->cut_values_.push_back(last);
+    p_cuts_->cut_values_.HostVector().push_back(last);
 
     // Ensure that every feature gets at least one quantile point
-    CHECK_LE(p_cuts_->cut_values_.size(), std::numeric_limits<uint32_t>::max());
-    auto cut_size = static_cast<uint32_t>(p_cuts_->cut_values_.size());
-    CHECK_GT(cut_size, p_cuts_->cut_ptrs_.back());
-    p_cuts_->cut_ptrs_.push_back(cut_size);
+    CHECK_LE(p_cuts_->cut_values_.HostVector().size(), std::numeric_limits<uint32_t>::max());
+    auto cut_size = static_cast<uint32_t>(p_cuts_->cut_values_.HostVector().size());
+    CHECK_GT(cut_size, p_cuts_->cut_ptrs_.HostVector().back());
+    p_cuts_->cut_ptrs_.HostVector().push_back(cut_size);
   }
   monitor_.Stop(__func__);
 }
 
-void GHistIndexMatrix::Init(DMatrix* p_fmat, int max_num_bins) {
-  cut.Build(p_fmat, max_num_bins);
+void GHistIndexMatrix::Init(DMatrix* p_fmat, int max_bins) {
+  cut.Build(p_fmat, max_bins);
+  max_num_bins = max_bins;
   const int32_t nthread = omp_get_max_threads();
   const uint32_t nbins = cut.Ptrs().back();
   hit_count.resize(nbins, 0);
   hit_count_tloc_.resize(nthread * nbins, 0);
 
-
+  this->p_fmat = p_fmat;
   size_t new_size = 1;
   for (const auto &batch : p_fmat->GetBatches<SparsePage>()) {
     new_size += batch.Size();
@@ -354,6 +468,8 @@ void GHistIndexMatrix::Init(DMatrix* p_fmat, int max_num_bins) {
 
   size_t rbegin = 0;
   size_t prev_sum = 0;
+  const bool isDense = p_fmat->IsDense();
+  this->isDense_ = isDense;
 
   for (const auto &batch : p_fmat->GetBatches<SparsePage>()) {
     // The number of threads is pegged to the batch size. If the OMP
@@ -400,25 +516,41 @@ void GHistIndexMatrix::Init(DMatrix* p_fmat, int max_num_bins) {
       }
     }
 
-    index.resize(row_ptr[rbegin + batch.Size()]);
+    const size_t n_offsets = cut.Ptrs().size() - 1;
+    const size_t n_index = row_ptr[rbegin + batch.Size()];
+    ResizeIndex(rbegin, batch, n_offsets, n_index, isDense);
 
     CHECK_GT(cut.Values().size(), 0U);
 
-    #pragma omp parallel for num_threads(batch_threads) schedule(static)
-    for (omp_ulong i = 0; i < batch.Size(); ++i) { // NOLINT(*)
-      const int tid = omp_get_thread_num();
-      size_t ibegin = row_ptr[rbegin + i];
-      size_t iend = row_ptr[rbegin + i + 1];
-      SparsePage::Inst inst = batch[i];
-
-      CHECK_EQ(ibegin + inst.size(), iend);
-      for (bst_uint j = 0; j < inst.size(); ++j) {
-        uint32_t idx = cut.SearchBin(inst[j]);
-
-        index[ibegin + j] = idx;
-        ++hit_count_tloc_[tid * nbins + idx];
+    uint32_t* offsets = nullptr;
+    if (isDense) {
+      index.ResizeOffset(n_offsets);
+      offsets = index.Offset();
+      for (size_t i = 0; i < n_offsets; ++i) {
+        offsets[i] = cut.Ptrs()[i];
       }
-      std::sort(index.begin() + ibegin, index.begin() + iend);
+    }
+
+    if (isDense) {
+      BinTypeSize curent_bin_size = index.GetBinTypeSize();
+      common::Span<const uint32_t> offsets_span = {offsets, n_offsets};
+      if (curent_bin_size == kUint8BinsTypeSize) {
+          common::Span<uint8_t> index_data_span = {index.data<uint8_t>(), n_index};
+          SetIndexDataForDense(index_data_span, batch_threads, batch, rbegin, offsets_span, nbins);
+      } else if (curent_bin_size == kUint16BinsTypeSize) {
+          common::Span<uint16_t> index_data_span = {index.data<uint16_t>(), n_index};
+          SetIndexDataForDense(index_data_span, batch_threads, batch, rbegin, offsets_span, nbins);
+      } else {
+          CHECK_EQ(curent_bin_size, kUint32BinsTypeSize);
+          common::Span<uint32_t> index_data_span = {index.data<uint32_t>(), n_index};
+          SetIndexDataForDense(index_data_span, batch_threads, batch, rbegin, offsets_span, nbins);
+      }
+
+    /* For sparse DMatrix we have to store index of feature for each bin
+       in index field to chose right offset. So offset is nullptr and index is not reduced */
+    } else {
+      common::Span<uint32_t> index_data_span = {index.data<uint32_t>(), n_index};
+      SetIndexDataForSparse(index_data_span, batch_threads, batch, rbegin, nbins);
     }
 
     #pragma omp parallel for num_threads(nthread) schedule(static)
@@ -434,13 +566,16 @@ void GHistIndexMatrix::Init(DMatrix* p_fmat, int max_num_bins) {
   }
 }
 
+template <typename BinIdxType>
 static size_t GetConflictCount(const std::vector<bool>& mark,
-                               const Column& column,
+                               const Column<BinIdxType>& column_input,
                                size_t max_cnt) {
   size_t ret = 0;
-  if (column.GetType() == xgboost::common::kDenseColumn) {
+  if (column_input.GetType() == xgboost::common::kDenseColumn) {
+    const DenseColumn<BinIdxType>& column
+      = static_cast<const DenseColumn<BinIdxType>& >(column_input);
     for (size_t i = 0; i < column.Size(); ++i) {
-      if (column.GetFeatureBinIdx(i) != std::numeric_limits<uint32_t>::max() && mark[i]) {
+      if ((!column.IsMissing(i)) && mark[i]) {
         ++ret;
         if (ret > max_cnt) {
           return max_cnt + 1;
@@ -448,6 +583,8 @@ static size_t GetConflictCount(const std::vector<bool>& mark,
       }
     }
   } else {
+    const SparseColumn<BinIdxType>& column
+      = static_cast<const SparseColumn<BinIdxType>& >(column_input);
     for (size_t i = 0; i < column.Size(); ++i) {
       if (mark[column.GetRowIdx(i)]) {
         ++ret;
@@ -460,22 +597,64 @@ static size_t GetConflictCount(const std::vector<bool>& mark,
   return ret;
 }
 
+template <typename BinIdxType>
 inline void
-MarkUsed(std::vector<bool>* p_mark, const Column& column) {
+MarkUsed(std::vector<bool>* p_mark, const Column<BinIdxType>& column_input) {
   std::vector<bool>& mark = *p_mark;
-  if (column.GetType() == xgboost::common::kDenseColumn) {
+  if (column_input.GetType() == xgboost::common::kDenseColumn) {
+    const DenseColumn<BinIdxType>& column
+      = static_cast<const DenseColumn<BinIdxType>& >(column_input);
     for (size_t i = 0; i < column.Size(); ++i) {
-      if (column.GetFeatureBinIdx(i) != std::numeric_limits<uint32_t>::max()) {
+      if (!column.IsMissing(i)) {
         mark[i] = true;
       }
     }
   } else {
+    const SparseColumn<BinIdxType>& column
+      = static_cast<const SparseColumn<BinIdxType>& >(column_input);
     for (size_t i = 0; i < column.Size(); ++i) {
       mark[column.GetRowIdx(i)] = true;
     }
   }
 }
 
+template <typename BinIdxType>
+inline void SetGroup(const unsigned fid, const Column<BinIdxType>& column,
+  const size_t max_conflict_cnt, const std::vector<size_t>& search_groups,
+  std::vector<size_t>* p_group_conflict_cnt,
+  std::vector<std::vector<bool>>* p_conflict_marks,
+  std::vector<std::vector<unsigned>>* p_groups,
+  std::vector<size_t>* p_group_nnz, const size_t cur_fid_nnz, const size_t nrow) {
+  bool need_new_group = true;
+  std::vector<size_t>& group_conflict_cnt = *p_group_conflict_cnt;
+  std::vector<std::vector<bool>>& conflict_marks = *p_conflict_marks;
+  std::vector<std::vector<unsigned>>& groups = *p_groups;
+  std::vector<size_t>& group_nnz = *p_group_nnz;
+
+  // examine each candidate group: is it okay to insert fid?
+  for (auto gid : search_groups) {
+    const size_t rest_max_cnt = max_conflict_cnt - group_conflict_cnt[gid];
+    const size_t cnt = GetConflictCount(conflict_marks[gid], column, rest_max_cnt);
+    if (cnt <= rest_max_cnt) {
+      need_new_group = false;
+      groups[gid].push_back(fid);
+      group_conflict_cnt[gid] += cnt;
+      group_nnz[gid] += cur_fid_nnz - cnt;
+      MarkUsed(&conflict_marks[gid], column);
+      break;
+    }
+  }
+  // create new group if necessary
+  if (need_new_group) {
+    groups.emplace_back();
+    groups.back().push_back(fid);
+    group_conflict_cnt.push_back(0);
+    conflict_marks.emplace_back(nrow, false);
+    MarkUsed(&conflict_marks.back(), column);
+    group_nnz.emplace_back(cur_fid_nnz);
+  }
+}
+
 inline std::vector<std::vector<unsigned>>
 FindGroups(const std::vector<unsigned>& feature_list,
            const std::vector<size_t>& feature_nnz,
@@ -495,10 +674,7 @@ FindGroups(const std::vector<unsigned>& feature_list,
     = static_cast<size_t>(param.max_conflict_rate * nrow);
 
   for (auto fid : feature_list) {
-    const Column& column = colmat.GetColumn(fid);
-
     const size_t cur_fid_nnz = feature_nnz[fid];
-    bool need_new_group = true;
 
     // randomly choose some of existing groups as candidates
     std::vector<size_t> search_groups;
@@ -512,31 +688,22 @@ FindGroups(const std::vector<unsigned>& feature_list,
       search_groups.resize(param.max_search_group);
     }
 
-    // examine each candidate group: is it okay to insert fid?
-    for (auto gid : search_groups) {
-      const size_t rest_max_cnt = max_conflict_cnt - group_conflict_cnt[gid];
-      const size_t cnt = GetConflictCount(conflict_marks[gid], column, rest_max_cnt);
-      if (cnt <= rest_max_cnt) {
-        need_new_group = false;
-        groups[gid].push_back(fid);
-        group_conflict_cnt[gid] += cnt;
-        group_nnz[gid] += cur_fid_nnz - cnt;
-        MarkUsed(&conflict_marks[gid], column);
-        break;
-      }
-    }
-
-    // create new group if necessary
-    if (need_new_group) {
-      groups.emplace_back();
-      groups.back().push_back(fid);
-      group_conflict_cnt.push_back(0);
-      conflict_marks.emplace_back(nrow, false);
-      MarkUsed(&conflict_marks.back(), column);
-      group_nnz.emplace_back(cur_fid_nnz);
+    BinTypeSize bins_type_size = colmat.GetTypeSize();
+    if (bins_type_size == kUint8BinsTypeSize) {
+        const auto column = colmat.GetColumn<uint8_t>(fid);
+        SetGroup(fid, *(column.get()), max_conflict_cnt, search_groups,
+                 &group_conflict_cnt, &conflict_marks, &groups, &group_nnz, cur_fid_nnz, nrow);
+    } else if (bins_type_size == kUint16BinsTypeSize) {
+        const auto column = colmat.GetColumn<uint16_t>(fid);
+        SetGroup(fid, *(column.get()), max_conflict_cnt, search_groups,
+                 &group_conflict_cnt, &conflict_marks, &groups, &group_nnz, cur_fid_nnz, nrow);
+    } else {
+        CHECK_EQ(bins_type_size, kUint32BinsTypeSize);
+        const auto column = colmat.GetColumn<uint32_t>(fid);
+        SetGroup(fid, *(column.get()), max_conflict_cnt, search_groups,
+                 &group_conflict_cnt, &conflict_marks, &groups, &group_nnz, cur_fid_nnz, nrow);
     }
   }
-
   return groups;
 }
 
@@ -618,6 +785,7 @@ void GHistIndexBlockMatrix::Init(const GHistIndexMatrix& gmat,
       }
     }
   }
+
   std::vector<std::vector<uint32_t>> index_temp(nblock);
   std::vector<std::vector<size_t>> row_ptr_temp(nblock);
   for (uint32_t block_id = 0; block_id < nblock; ++block_id) {
@@ -660,7 +828,7 @@ void GHistIndexBlockMatrix::Init(const GHistIndexMatrix& gmat,
 }
 
 /*!
- * \brief fill a histogram by zeroes
+ * \brief fill a histogram by zeros in range [begin, end)
  */
 void InitilizeHistByZeroes(GHistRow hist, size_t begin, size_t end) {
 #if defined(XGBOOST_STRICT_R_MODE) && XGBOOST_STRICT_R_MODE == 1
@@ -711,40 +879,171 @@ void SubtractionHist(GHistRow dst, const GHistRow src1, const GHistRow src2,
   }
 }
 
+struct Prefetch {
+ public:
+  static constexpr size_t kCacheLineSize = 64;
+  static constexpr size_t kPrefetchOffset = 10;
+
+ private:
+  static constexpr size_t kNoPrefetchSize =
+      kPrefetchOffset + kCacheLineSize /
+      sizeof(decltype(GHistIndexMatrix::row_ptr)::value_type);
+
+ public:
+  static size_t NoPrefetchSize(size_t rows) {
+    return std::min(rows, kNoPrefetchSize);
+  }
+
+  template <typename T>
+  static constexpr size_t GetPrefetchStep() {
+    return Prefetch::kCacheLineSize / sizeof(T);
+  }
+};
+
+constexpr size_t Prefetch::kNoPrefetchSize;
+
+
+template<typename FPType, bool do_prefetch, typename BinIdxType>
+void BuildHistDenseKernel(const std::vector<GradientPair>& gpair,
+                          const RowSetCollection::Elem row_indices,
+                          const GHistIndexMatrix& gmat,
+                          const size_t n_features,
+                          GHistRow hist) {
+  const size_t size = row_indices.Size();
+  const size_t* rid = row_indices.begin;
+  const float* pgh = reinterpret_cast<const float*>(gpair.data());
+  const BinIdxType* gradient_index = gmat.index.data<BinIdxType>();
+  const uint32_t* offsets = gmat.index.Offset();
+  FPType* hist_data = reinterpret_cast<FPType*>(hist.data());
+  const uint32_t two {2};  // Each element from 'gpair' and 'hist' contains
+                           // 2 FP values: gradient and hessian.
+                           // So we need to multiply each row-index/bin-index by 2
+                           // to work with gradient pairs as a singe row FP array
+
+  for (size_t i = 0; i < size; ++i) {
+    const size_t icol_start = rid[i] * n_features;
+    const size_t idx_gh = two * rid[i];
+
+    if (do_prefetch) {
+      const size_t icol_start_prefetch = rid[i + Prefetch::kPrefetchOffset] * n_features;
+
+      PREFETCH_READ_T0(pgh + two * rid[i + Prefetch::kPrefetchOffset]);
+      for (size_t j = icol_start_prefetch; j < icol_start_prefetch + n_features;
+           j += Prefetch::GetPrefetchStep<BinIdxType>()) {
+        PREFETCH_READ_T0(gradient_index + j);
+      }
+    }
+    const BinIdxType* gr_index_local = gradient_index + icol_start;
+    for (size_t j = 0; j < n_features; ++j) {
+      const uint32_t idx_bin = two * (static_cast<uint32_t>(gr_index_local[j]) +
+                                      offsets[j]);
+
+      hist_data[idx_bin]   += pgh[idx_gh];
+      hist_data[idx_bin+1] += pgh[idx_gh+1];
+    }
+  }
+}
+
+template<typename FPType, bool do_prefetch>
+void BuildHistSparseKernel(const std::vector<GradientPair>& gpair,
+                           const RowSetCollection::Elem row_indices,
+                           const GHistIndexMatrix& gmat,
+                           GHistRow hist) {
+  const size_t size = row_indices.Size();
+  const size_t* rid = row_indices.begin;
+  const float* pgh = reinterpret_cast<const float*>(gpair.data());
+  const uint32_t* gradient_index = gmat.index.data<uint32_t>();
+  const size_t* row_ptr =  gmat.row_ptr.data();
+  FPType* hist_data = reinterpret_cast<FPType*>(hist.data());
+  const uint32_t two {2};  // Each element from 'gpair' and 'hist' contains
+                           // 2 FP values: gradient and hessian.
+                           // So we need to multiply each row-index/bin-index by 2
+                           // to work with gradient pairs as a singe row FP array
+
+  for (size_t i = 0; i < size; ++i) {
+    const size_t icol_start = row_ptr[rid[i]];
+    const size_t icol_end = row_ptr[rid[i]+1];
+    const size_t idx_gh = two * rid[i];
+
+    if (do_prefetch) {
+      const size_t icol_start_prftch = row_ptr[rid[i+Prefetch::kPrefetchOffset]];
+      const size_t icol_end_prefect = row_ptr[rid[i+Prefetch::kPrefetchOffset]+1];
+
+      PREFETCH_READ_T0(pgh + two * rid[i + Prefetch::kPrefetchOffset]);
+      for (size_t j = icol_start_prftch; j < icol_end_prefect;
+        j+=Prefetch::GetPrefetchStep<uint32_t>()) {
+        PREFETCH_READ_T0(gradient_index + j);
+      }
+    }
+    for (size_t j = icol_start; j < icol_end; ++j) {
+      const uint32_t idx_bin = two * gradient_index[j];
+      hist_data[idx_bin]   += pgh[idx_gh];
+      hist_data[idx_bin+1] += pgh[idx_gh+1];
+    }
+  }
+}
+
+
+template<typename FPType, bool do_prefetch, typename BinIdxType>
+void BuildHistDispatchKernel(const std::vector<GradientPair>& gpair,
+                     const RowSetCollection::Elem row_indices,
+                     const GHistIndexMatrix& gmat, GHistRow hist, bool isDense) {
+  if (isDense) {
+    const size_t* row_ptr =  gmat.row_ptr.data();
+    const size_t n_features = row_ptr[row_indices.begin[0]+1] - row_ptr[row_indices.begin[0]];
+    BuildHistDenseKernel<FPType, do_prefetch, BinIdxType>(gpair, row_indices,
+                                                       gmat, n_features, hist);
+  } else {
+    BuildHistSparseKernel<FPType, do_prefetch>(gpair, row_indices,
+                                                        gmat, hist);
+  }
+}
+
+template<typename FPType, bool do_prefetch>
+void BuildHistKernel(const std::vector<GradientPair>& gpair,
+                     const RowSetCollection::Elem row_indices,
+                     const GHistIndexMatrix& gmat, const bool isDense, GHistRow hist) {
+  const bool is_dense = row_indices.Size() && isDense;
+  switch (gmat.index.GetBinTypeSize()) {
+    case kUint8BinsTypeSize:
+      BuildHistDispatchKernel<FPType, do_prefetch, uint8_t>(gpair, row_indices,
+                                                            gmat, hist, is_dense);
+      break;
+    case kUint16BinsTypeSize:
+      BuildHistDispatchKernel<FPType, do_prefetch, uint16_t>(gpair, row_indices,
+                                                             gmat, hist, is_dense);
+      break;
+    case kUint32BinsTypeSize:
+      BuildHistDispatchKernel<FPType, do_prefetch, uint32_t>(gpair, row_indices,
+                                                             gmat, hist, is_dense);
+      break;
+    default:
+      CHECK(false);  // no default behavior
+  }
+}
 
 void GHistBuilder::BuildHist(const std::vector<GradientPair>& gpair,
                              const RowSetCollection::Elem row_indices,
                              const GHistIndexMatrix& gmat,
-                             GHistRow hist) {
-  const size_t* rid =  row_indices.begin;
+                             GHistRow hist,
+                             bool isDense) {
+  using FPType = decltype(tree::GradStats::sum_grad);
   const size_t nrows = row_indices.Size();
-  const uint32_t* index = gmat.index.data();
-  const size_t* row_ptr =  gmat.row_ptr.data();
-  const float* pgh = reinterpret_cast<const float*>(gpair.data());
+  const size_t no_prefetch_size = Prefetch::NoPrefetchSize(nrows);
 
-  double* hist_data = reinterpret_cast<double*>(hist.data());
+  // if need to work with all rows from bin-matrix (e.g. root node)
+  const bool contiguousBlock = (row_indices.begin[nrows - 1] - row_indices.begin[0]) == (nrows - 1);
 
-  const size_t cache_line_size = 64;
-  const size_t prefetch_offset = 10;
-  size_t no_prefetch_size = prefetch_offset + cache_line_size/sizeof(*rid);
-  no_prefetch_size = no_prefetch_size > nrows ? nrows : no_prefetch_size;
+  if (contiguousBlock) {
+    // contiguous memory access, built-in HW prefetching is enough
+    BuildHistKernel<FPType, false>(gpair, row_indices, gmat, isDense, hist);
+  } else {
+    const RowSetCollection::Elem span1(row_indices.begin, row_indices.end - no_prefetch_size);
+    const RowSetCollection::Elem span2(row_indices.end - no_prefetch_size, row_indices.end);
 
-  for (size_t i = 0; i < nrows; ++i) {
-    const size_t icol_start = row_ptr[rid[i]];
-    const size_t icol_end = row_ptr[rid[i]+1];
-
-    if (i < nrows - no_prefetch_size) {
-      PREFETCH_READ_T0(row_ptr + rid[i + prefetch_offset]);
-      PREFETCH_READ_T0(pgh + 2*rid[i + prefetch_offset]);
-    }
-
-    for (size_t j = icol_start; j < icol_end; ++j) {
-      const uint32_t idx_bin = 2*index[j];
-      const size_t idx_gh = 2*rid[i];
-
-      hist_data[idx_bin] += pgh[idx_gh];
-      hist_data[idx_bin+1] += pgh[idx_gh+1];
-    }
+    BuildHistKernel<FPType, true>(gpair, span1, gmat, isDense, hist);
+    // no prefetching to avoid loading extra memory
+    BuildHistKernel<FPType, false>(gpair, span2, gmat, isDense, hist);
   }
 }
 
@@ -756,7 +1055,6 @@ void GHistBuilder::BuildBlockHist(const std::vector<GradientPair>& gpair,
   const size_t nblock = gmatb.GetNumBlock();
   const size_t nrows = row_indices.end - row_indices.begin;
   const size_t rest = nrows % kUnroll;
-
 #if defined(_OPENMP)
   const auto nthread = static_cast<bst_omp_uint>(this->nthread_);  // NOLINT
 #endif  // defined(_OPENMP)
