@@ -1,5 +1,5 @@
 /*!
- * Copyright 2014 by Contributors
+ * Copyright 2014~2020 by Contributors
  * \file simple_dmatrix.cc
  * \brief the input data structure for gradient boosting
  * \author Tianqi Chen
@@ -8,41 +8,46 @@
 #include <xgboost/data.h>
 #include "./simple_batch_iterator.h"
 #include "../common/random.h"
+#include "adapter.h"
 
 namespace xgboost {
 namespace data {
-MetaInfo& SimpleDMatrix::Info() { return source_->info; }
+MetaInfo& SimpleDMatrix::Info() { return info_; }
 
-const MetaInfo& SimpleDMatrix::Info() const { return source_->info; }
+const MetaInfo& SimpleDMatrix::Info() const { return info_; }
 
-float SimpleDMatrix::GetColDensity(size_t cidx) {
-  size_t column_size = 0;
-  // Use whatever version of column batches already exists
-  if (sorted_column_page_) {
-    auto batch = this->GetBatches<SortedCSCPage>();
-    column_size = (*batch.begin())[cidx].size();
-  } else {
-    auto batch = this->GetBatches<CSCPage>();
-    column_size = (*batch.begin())[cidx].size();
+DMatrix* SimpleDMatrix::Slice(common::Span<int32_t const> ridxs) {
+  auto out = new SimpleDMatrix;
+  SparsePage& out_page = out->sparse_page_;
+  for (auto const &page : this->GetBatches<SparsePage>()) {
+    page.data.HostVector();
+    page.offset.HostVector();
+    auto& h_data = out_page.data.HostVector();
+    auto& h_offset = out_page.offset.HostVector();
+    size_t rptr{0};
+    for (auto ridx : ridxs) {
+      auto inst = page[ridx];
+      rptr += inst.size();
+      std::copy(inst.begin(), inst.end(), std::back_inserter(h_data));
+      h_offset.emplace_back(rptr);
+    }
+    out->Info() = this->Info().Slice(ridxs);
+    out->Info().num_nonzero_ = h_offset.back();
   }
-
-  size_t nmiss = this->Info().num_row_ - column_size;
-  return 1.0f - (static_cast<float>(nmiss)) / this->Info().num_row_;
+  return out;
 }
 
 BatchSet<SparsePage> SimpleDMatrix::GetRowBatches() {
   // since csr is the default data structure so `source_` is always available.
-  auto cast = dynamic_cast<SimpleCSRSource*>(source_.get());
   auto begin_iter = BatchIterator<SparsePage>(
-      new SimpleBatchIteratorImpl<SparsePage>(&(cast->page_)));
+      new SimpleBatchIteratorImpl<SparsePage>(&sparse_page_));
   return BatchSet<SparsePage>(begin_iter);
 }
 
 BatchSet<CSCPage> SimpleDMatrix::GetColumnBatches() {
   // column page doesn't exist, generate it
   if (!column_page_) {
-    auto const& page = dynamic_cast<SimpleCSRSource*>(source_.get())->page_;
-    column_page_.reset(new CSCPage(page.GetTranspose(source_->info.num_col_)));
+    column_page_.reset(new CSCPage(sparse_page_.GetTranspose(info_.num_col_)));
   }
   auto begin_iter =
       BatchIterator<CSCPage>(new SimpleBatchIteratorImpl<CSCPage>(column_page_.get()));
@@ -52,9 +57,8 @@ BatchSet<CSCPage> SimpleDMatrix::GetColumnBatches() {
 BatchSet<SortedCSCPage> SimpleDMatrix::GetSortedColumnBatches() {
   // Sorted column page doesn't exist, generate it
   if (!sorted_column_page_) {
-    auto const& page = dynamic_cast<SimpleCSRSource*>(source_.get())->page_;
     sorted_column_page_.reset(
-        new SortedCSCPage(page.GetTranspose(source_->info.num_col_)));
+        new SortedCSCPage(sparse_page_.GetTranspose(info_.num_col_)));
     sorted_column_page_->SortRows();
   }
   auto begin_iter = BatchIterator<SortedCSCPage>(
@@ -63,19 +67,21 @@ BatchSet<SortedCSCPage> SimpleDMatrix::GetSortedColumnBatches() {
 }
 
 BatchSet<EllpackPage> SimpleDMatrix::GetEllpackBatches(const BatchParam& param) {
-  CHECK_GE(param.gpu_id, 0);
-  CHECK_GE(param.max_bin, 2);
   // ELLPACK page doesn't exist, generate it
-  if (!ellpack_page_) {
+  if (!(batch_param_ != BatchParam{})) {
+    CHECK(param != BatchParam{}) << "Batch parameter is not initialized.";
+  }
+  if (!ellpack_page_  || (batch_param_ != param && param != BatchParam{})) {
+    CHECK_GE(param.gpu_id, 0);
+    CHECK_GE(param.max_bin, 2);
     ellpack_page_.reset(new EllpackPage(this, param));
+    batch_param_ = param;
   }
   auto begin_iter =
       BatchIterator<EllpackPage>(new SimpleBatchIteratorImpl<EllpackPage>(ellpack_page_.get()));
   return BatchSet<EllpackPage>(begin_iter);
 }
 
-bool SimpleDMatrix::SingleColBlock() const { return true; }
-
 template <typename AdapterT>
 SimpleDMatrix::SimpleDMatrix(AdapterT* adapter, float missing, int nthread) {
   // Set number of threads but keep old value so we can reset it after
@@ -84,35 +90,33 @@ SimpleDMatrix::SimpleDMatrix(AdapterT* adapter, float missing, int nthread) {
   int nthread_original = omp_get_max_threads();
   omp_set_num_threads(nthread);
 
-  source_.reset(new SimpleCSRSource());
-  SimpleCSRSource& mat = *reinterpret_cast<SimpleCSRSource*>(source_.get());
   std::vector<uint64_t> qids;
   uint64_t default_max = std::numeric_limits<uint64_t>::max();
   uint64_t last_group_id = default_max;
   bst_uint group_size = 0;
-  auto& offset_vec = mat.page_.offset.HostVector();
-  auto& data_vec = mat.page_.data.HostVector();
+  auto& offset_vec = sparse_page_.offset.HostVector();
+  auto& data_vec = sparse_page_.data.HostVector();
   uint64_t inferred_num_columns = 0;
 
   adapter->BeforeFirst();
   // Iterate over batches of input data
   while (adapter->Next()) {
     auto& batch = adapter->Value();
-    auto batch_max_columns = mat.page_.Push(batch, missing, nthread);
+    auto batch_max_columns = sparse_page_.Push(batch, missing, nthread);
     inferred_num_columns = std::max(batch_max_columns, inferred_num_columns);
     // Append meta information if available
     if (batch.Labels() != nullptr) {
-      auto& labels = mat.info.labels_.HostVector();
+      auto& labels = info_.labels_.HostVector();
       labels.insert(labels.end(), batch.Labels(),
                     batch.Labels() + batch.Size());
     }
     if (batch.Weights() != nullptr) {
-      auto& weights = mat.info.weights_.HostVector();
+      auto& weights = info_.weights_.HostVector();
       weights.insert(weights.end(), batch.Weights(),
                      batch.Weights() + batch.Size());
     }
     if (batch.BaseMargin() != nullptr) {
-      auto& base_margin = mat.info.base_margin_.HostVector();
+      auto& base_margin = info_.base_margin_.HostVector();
       base_margin.insert(base_margin.end(), batch.BaseMargin(),
                      batch.BaseMargin() + batch.Size());
     }
@@ -122,7 +126,7 @@ SimpleDMatrix::SimpleDMatrix(AdapterT* adapter, float missing, int nthread) {
       for (size_t i = 0; i < batch.Size(); ++i) {
         const uint64_t cur_group_id = batch.Qid()[i];
         if (last_group_id == default_max || last_group_id != cur_group_id) {
-          mat.info.group_ptr_.push_back(group_size);
+          info_.group_ptr_.push_back(group_size);
         }
         last_group_id = cur_group_id;
         ++group_size;
@@ -131,22 +135,23 @@ SimpleDMatrix::SimpleDMatrix(AdapterT* adapter, float missing, int nthread) {
   }
 
   if (last_group_id != default_max) {
-    if (group_size > mat.info.group_ptr_.back()) {
-      mat.info.group_ptr_.push_back(group_size);
+    if (group_size > info_.group_ptr_.back()) {
+      info_.group_ptr_.push_back(group_size);
     }
   }
 
   // Deal with empty rows/columns if necessary
   if (adapter->NumColumns() == kAdapterUnknownSize) {
-    mat.info.num_col_ = inferred_num_columns;
+    info_.num_col_ = inferred_num_columns;
   } else {
-    mat.info.num_col_ = adapter->NumColumns();
+    info_.num_col_ = adapter->NumColumns();
   }
+
   // Synchronise worker columns
-  rabit::Allreduce<rabit::op::Max>(&mat.info.num_col_, 1);
+  rabit::Allreduce<rabit::op::Max>(&info_.num_col_, 1);
 
   if (adapter->NumRows() == kAdapterUnknownSize) {
-    mat.info.num_row_ = offset_vec.size() - 1;
+    info_.num_row_ = offset_vec.size() - 1;
   } else {
     if (offset_vec.empty()) {
       offset_vec.emplace_back(0);
@@ -155,12 +160,31 @@ SimpleDMatrix::SimpleDMatrix(AdapterT* adapter, float missing, int nthread) {
     while (offset_vec.size() - 1 < adapter->NumRows()) {
       offset_vec.emplace_back(offset_vec.back());
     }
-    mat.info.num_row_ = adapter->NumRows();
+    info_.num_row_ = adapter->NumRows();
   }
-  mat.info.num_nonzero_ = data_vec.size();
+  info_.num_nonzero_ = data_vec.size();
   omp_set_num_threads(nthread_original);
 }
 
+SimpleDMatrix::SimpleDMatrix(dmlc::Stream* in_stream) {
+  int tmagic;
+  CHECK(in_stream->Read(&tmagic, sizeof(tmagic)) == sizeof(tmagic))
+      << "invalid input file format";
+  CHECK_EQ(tmagic, kMagic) << "invalid format, magic number mismatch";
+  info_.LoadBinary(in_stream);
+  in_stream->Read(&sparse_page_.offset.HostVector());
+  in_stream->Read(&sparse_page_.data.HostVector());
+}
+
+void SimpleDMatrix::SaveToLocalFile(const std::string& fname) {
+    std::unique_ptr<dmlc::Stream> fo(dmlc::Stream::Create(fname.c_str(), "w"));
+    int tmagic = kMagic;
+    fo->Write(&tmagic, sizeof(tmagic));
+    info_.SaveBinary(fo.get());
+    fo->Write(sparse_page_.offset.HostVector());
+    fo->Write(sparse_page_.data.HostVector());
+}
+
 template SimpleDMatrix::SimpleDMatrix(DenseAdapter* adapter, float missing,
                                      int nthread);
 template SimpleDMatrix::SimpleDMatrix(CSRAdapter* adapter, float missing,
@@ -171,7 +195,7 @@ template SimpleDMatrix::SimpleDMatrix(DataTableAdapter* adapter, float missing,
                                      int nthread);
 template SimpleDMatrix::SimpleDMatrix(FileAdapter* adapter, float missing,
                                      int nthread);
-template SimpleDMatrix::SimpleDMatrix(DMatrixSliceAdapter* adapter, float missing,
+template SimpleDMatrix::SimpleDMatrix(IteratorAdapter* adapter, float missing,
                                      int nthread);
 }  // namespace data
 }  // namespace xgboost
