@@ -6,17 +6,25 @@
  */
 #include <dmlc/io.h>
 #include <dmlc/parameter.h>
+#include <dmlc/thread_local.h>
 
+#include <atomic>
+#include <mutex>
 #include <algorithm>
 #include <iomanip>
 #include <limits>
+#include <memory>
 #include <sstream>
 #include <string>
 #include <stack>
 #include <utility>
 #include <vector>
 
+#include "dmlc/any.h"
 #include "xgboost/base.h"
+#include "xgboost/data.h"
+#include "xgboost/model.h"
+#include "xgboost/predictor.h"
 #include "xgboost/feature_map.h"
 #include "xgboost/gbm.h"
 #include "xgboost/generic_parameters.h"
@@ -189,18 +197,42 @@ void GenericParameter::ConfigureGpuId(bool require_gpu) {
 #endif  // defined(XGBOOST_USE_CUDA)
 }
 
-/*!
- * \brief learner that performs gradient boosting for a specific objective
- * function. It does training and prediction.
- */
-class LearnerImpl : public Learner {
+using XGBAPIThreadLocalStore =
+    dmlc::ThreadLocalStore<std::map<Learner const *, XGBAPIThreadLocalEntry>>;
+
+class LearnerConfiguration : public Learner {
+ protected:
+  static std::string const kEvalMetric;  // NOLINT
+
+ protected:
+  PredictionContainer cache_;
+
+ protected:
+  std::atomic<bool> need_configuration_;
+  std::map<std::string, std::string> cfg_;
+  // Stores information like best-iteration for early stopping.
+  std::map<std::string, std::string> attributes_;
+  common::Monitor monitor_;
+  LearnerModelParamLegacy mparam_;
+  LearnerModelParam learner_model_param_;
+  LearnerTrainParam tparam_;
+  std::vector<std::string> metric_names_;
+  std::mutex config_lock_;
+
  public:
-  explicit LearnerImpl(std::vector<std::shared_ptr<DMatrix> >  cache)
-      : need_configuration_{true}, cache_(std::move(cache)) {
+  explicit LearnerConfiguration(std::vector<std::shared_ptr<DMatrix> > cache)
+      : need_configuration_{true} {
     monitor_.Init("Learner");
+    for (std::shared_ptr<DMatrix> const& d : cache) {
+      cache_.Cache(d, GenericParameter::kCpuId);
+    }
   }
   // Configuration before data is known.
+
   void Configure() override {
+    // Varient of double checked lock
+    if (!this->need_configuration_) { return; }
+    std::lock_guard<std::mutex> gard(config_lock_);
     if (!this->need_configuration_) { return; }
 
     monitor_.Start("Configure");
@@ -263,6 +295,126 @@ class LearnerImpl : public Learner {
     monitor_.Stop("Configure");
   }
 
+  void LoadConfig(Json const& in) override {
+    CHECK(IsA<Object>(in));
+    Version::Load(in, true);
+
+    auto const& learner_parameters = get<Object>(in["learner"]);
+    FromJson(learner_parameters.at("learner_train_param"), &tparam_);
+
+    auto const& gradient_booster = learner_parameters.at("gradient_booster");
+
+    auto const& objective_fn = learner_parameters.at("objective");
+    if (!obj_) {
+      obj_.reset(ObjFunction::Create(tparam_.objective, &generic_parameters_));
+    }
+    obj_->LoadConfig(objective_fn);
+
+    tparam_.booster = get<String>(gradient_booster["name"]);
+    if (!gbm_) {
+      gbm_.reset(GradientBooster::Create(tparam_.booster,
+                                         &generic_parameters_, &learner_model_param_));
+    }
+    gbm_->LoadConfig(gradient_booster);
+
+    auto const& j_metrics = learner_parameters.at("metrics");
+    auto n_metrics = get<Array const>(j_metrics).size();
+    metric_names_.resize(n_metrics);
+    metrics_.resize(n_metrics);
+    for (size_t i = 0; i < n_metrics; ++i) {
+      metric_names_[i]= get<String>(j_metrics[i]);
+      metrics_[i] = std::unique_ptr<Metric>(
+          Metric::Create(metric_names_[i], &generic_parameters_));
+    }
+
+    FromJson(learner_parameters.at("generic_param"), &generic_parameters_);
+    // make sure the GPU ID is valid in new environment before start running configure.
+    generic_parameters_.ConfigureGpuId(false);
+
+    this->need_configuration_ = true;
+  }
+
+  void SaveConfig(Json* p_out) const override {
+    CHECK(!this->need_configuration_) << "Call Configure before saving model.";
+    Version::Save(p_out);
+    Json& out { *p_out };
+    // parameters
+    out["learner"] = Object();
+    auto& learner_parameters = out["learner"];
+
+    learner_parameters["learner_train_param"] = ToJson(tparam_);
+    learner_parameters["learner_model_param"] = mparam_.ToJson();
+    learner_parameters["gradient_booster"] = Object();
+    auto& gradient_booster = learner_parameters["gradient_booster"];
+    gbm_->SaveConfig(&gradient_booster);
+
+    learner_parameters["objective"] = Object();
+    auto& objective_fn = learner_parameters["objective"];
+    obj_->SaveConfig(&objective_fn);
+
+    std::vector<Json> metrics(metrics_.size());
+    for (size_t i = 0; i < metrics_.size(); ++i) {
+      metrics[i] = String(metrics_[i]->Name());
+    }
+    learner_parameters["metrics"] = Array(std::move(metrics));
+
+    learner_parameters["generic_param"] = ToJson(generic_parameters_);
+  }
+
+  void SetParam(const std::string& key, const std::string& value) override {
+    this->need_configuration_ = true;
+    if (key == kEvalMetric) {
+      if (std::find(metric_names_.cbegin(), metric_names_.cend(),
+                    value) == metric_names_.cend()) {
+        metric_names_.emplace_back(value);
+      }
+    } else {
+      cfg_[key] = value;
+    }
+  }
+  // Short hand for setting multiple parameters
+  void SetParams(std::vector<std::pair<std::string, std::string>> const& args) override {
+    for (auto const& kv : args) {
+      this->SetParam(kv.first, kv.second);
+    }
+  }
+
+  void SetAttr(const std::string& key, const std::string& value) override {
+    attributes_[key] = value;
+    mparam_.contain_extra_attrs = 1;
+  }
+
+  bool GetAttr(const std::string& key, std::string* out) const override {
+    auto it = attributes_.find(key);
+    if (it == attributes_.end()) return false;
+    *out = it->second;
+    return true;
+  }
+
+  bool DelAttr(const std::string& key) override {
+    auto it = attributes_.find(key);
+    if (it == attributes_.end()) { return false; }
+    attributes_.erase(it);
+    return true;
+  }
+
+  std::vector<std::string> GetAttrNames() const override {
+    std::vector<std::string> out;
+    for (auto const& kv : attributes_) {
+      out.emplace_back(kv.first);
+    }
+    return out;
+  }
+
+  const std::map<std::string, std::string>& GetConfigurationArguments() const override {
+    return cfg_;
+  }
+
+  GenericParameter const& GetGenericParameter() const override {
+    return generic_parameters_;
+  }
+
+ private:
   void ValidateParameters() {
     Json config { Object() };
     this->SaveConfig(&config);
@@ -329,18 +481,98 @@ class LearnerImpl : public Learner {
     }
   }
 
-  void CheckDataSplitMode() {
-    if (rabit::IsDistributed()) {
-      CHECK(tparam_.dsplit != DataSplitMode::kAuto)
-        << "Precondition violated; dsplit cannot be 'auto' in distributed mode";
-      if (tparam_.dsplit == DataSplitMode::kCol) {
-        // 'distcol' updater hidden until it becomes functional again
-        // See discussion at https://github.com/dmlc/xgboost/issues/1832
-        LOG(FATAL) << "Column-wise data split is currently not supported.";
+  void ConfigureNumFeatures() {
+    // Compute number of global features if parameter not already set
+    if (mparam_.num_feature == 0) {
+      // TODO(hcho3): Change num_feature to 64-bit integer
+      unsigned num_feature = 0;
+      for (auto& matrix : cache_.Container()) {
+        CHECK(matrix.first);
+        CHECK(!matrix.second.ref.expired());
+        const uint64_t num_col = matrix.first->Info().num_col_;
+        CHECK_LE(num_col,
+                 static_cast<uint64_t>(std::numeric_limits<unsigned>::max()))
+            << "Unfortunately, XGBoost does not support data matrices with "
+            << std::numeric_limits<unsigned>::max() << " features or greater";
+        num_feature = std::max(num_feature, static_cast<uint32_t>(num_col));
+      }
+
+      rabit::Allreduce<rabit::op::Max>(&num_feature, 1, nullptr, nullptr,
+                                       "num_feature");
+      if (num_feature > mparam_.num_feature) {
+        mparam_.num_feature = num_feature;
       }
     }
+    CHECK_NE(mparam_.num_feature, 0)
+        << "0 feature is supplied.  Are you using raw Booster interface?";
+    // Remove these once binary IO is gone.
+    cfg_["num_feature"] = common::ToString(mparam_.num_feature);
+    cfg_["num_class"] = common::ToString(mparam_.num_class);
   }
 
+  void ConfigureGBM(LearnerTrainParam const& old, Args const& args) {
+    if (gbm_ == nullptr || old.booster != tparam_.booster) {
+      gbm_.reset(GradientBooster::Create(tparam_.booster, &generic_parameters_,
+                                         &learner_model_param_));
+    }
+    gbm_->Configure(args);
+  }
+
+  void ConfigureObjective(LearnerTrainParam const& old, Args* p_args) {
+    // Once binary IO is gone, NONE of these config is useful.
+    if (cfg_.find("num_class") != cfg_.cend() && cfg_.at("num_class") != "0" &&
+        tparam_.objective != "multi:softprob") {
+      cfg_["num_output_group"] = cfg_["num_class"];
+      if (atoi(cfg_["num_class"].c_str()) > 1 && cfg_.count("objective") == 0) {
+        tparam_.objective = "multi:softmax";
+      }
+    }
+
+    if (cfg_.find("max_delta_step") == cfg_.cend() &&
+        cfg_.find("objective") != cfg_.cend() &&
+        tparam_.objective == "count:poisson") {
+      // max_delta_step is a duplicated parameter in Poisson regression and tree param.
+      // Rename one of them once binary IO is gone.
+      cfg_["max_delta_step"] = kMaxDeltaStepDefaultValue;
+    }
+    if (obj_ == nullptr || tparam_.objective != old.objective) {
+      obj_.reset(ObjFunction::Create(tparam_.objective, &generic_parameters_));
+    }
+    auto& args = *p_args;
+    args = {cfg_.cbegin(), cfg_.cend()};  // renew
+    obj_->Configure(args);
+  }
+
+  void ConfigureMetrics(Args const& args) {
+    for (auto const& name : metric_names_) {
+      auto DupCheck = [&name](std::unique_ptr<Metric> const& m) {
+                        return m->Name() != name;
+                      };
+      if (std::all_of(metrics_.begin(), metrics_.end(), DupCheck)) {
+        metrics_.emplace_back(std::unique_ptr<Metric>(Metric::Create(name, &generic_parameters_)));
+        mparam_.contain_eval_metrics = 1;
+      }
+    }
+    for (auto& p_metric : metrics_) {
+      p_metric->Configure(args);
+    }
+  }
+};
+
+std::string const LearnerConfiguration::kEvalMetric {"eval_metric"};  // NOLINT
+
+class LearnerIO : public LearnerConfiguration {
+ private:
+  std::set<std::string> saved_configs_ = {"num_round"};
+  // Used to identify the offset of JSON string when
+  // `enable_experimental_json_serialization' is set to false.  Will be removed once JSON
+  // takes over.
+  std::string const serialisation_header_ { u8"CONFIG-offset:" };
+
+ public:
+  explicit LearnerIO(std::vector<std::shared_ptr<DMatrix> > cache) :
+      LearnerConfiguration{cache} {}
+
   void LoadModel(Json const& in) override {
     CHECK(IsA<Object>(in));
     Version::Load(in, false);
@@ -358,8 +590,7 @@ class LearnerImpl : public Learner {
     name = get<String>(gradient_booster["name"]);
     tparam_.UpdateAllowUnknown(Args{{"booster", name}});
     gbm_.reset(GradientBooster::Create(tparam_.booster,
-                                       &generic_parameters_, &learner_model_param_,
-                                       cache_));
+                                       &generic_parameters_, &learner_model_param_));
     gbm_->LoadModel(gradient_booster);
 
     auto const& j_attributes = get<Object const>(learner.at("attributes"));
@@ -394,74 +625,6 @@ class LearnerImpl : public Learner {
       learner["attributes"][kv.first] = String(kv.second);
     }
   }
-
-  void LoadConfig(Json const& in) override {
-    CHECK(IsA<Object>(in));
-    Version::Load(in, true);
-
-    auto const& learner_parameters = get<Object>(in["learner"]);
-    fromJson(learner_parameters.at("learner_train_param"), &tparam_);
-
-    auto const& gradient_booster = learner_parameters.at("gradient_booster");
-
-    auto const& objective_fn = learner_parameters.at("objective");
-    if (!obj_) {
-      obj_.reset(ObjFunction::Create(tparam_.objective, &generic_parameters_));
-    }
-    obj_->LoadConfig(objective_fn);
-
-    tparam_.booster = get<String>(gradient_booster["name"]);
-    if (!gbm_) {
-      gbm_.reset(GradientBooster::Create(tparam_.booster,
-                                         &generic_parameters_, &learner_model_param_,
-                                         cache_));
-    }
-    gbm_->LoadConfig(gradient_booster);
-
-    auto const& j_metrics = learner_parameters.at("metrics");
-    auto n_metrics = get<Array const>(j_metrics).size();
-    metric_names_.resize(n_metrics);
-    metrics_.resize(n_metrics);
-    for (size_t i = 0; i < n_metrics; ++i) {
-      metric_names_[i]= get<String>(j_metrics[i]);
-      metrics_[i] = std::unique_ptr<Metric>(
-          Metric::Create(metric_names_[i], &generic_parameters_));
-    }
-
-    fromJson(learner_parameters.at("generic_param"), &generic_parameters_);
-    // make sure the GPU ID is valid in new environment before start running configure.
-    generic_parameters_.ConfigureGpuId(false);
-
-    this->need_configuration_ = true;
-  }
-
-  void SaveConfig(Json* p_out) const override {
-    CHECK(!this->need_configuration_) << "Call Configure before saving model.";
-    Version::Save(p_out);
-    Json& out { *p_out };
-    // parameters
-    out["learner"] = Object();
-    auto& learner_parameters = out["learner"];
-
-    learner_parameters["learner_train_param"] = toJson(tparam_);
-    learner_parameters["learner_model_param"] = mparam_.ToJson();
-    learner_parameters["gradient_booster"] = Object();
-    auto& gradient_booster = learner_parameters["gradient_booster"];
-    gbm_->SaveConfig(&gradient_booster);
-
-    learner_parameters["objective"] = Object();
-    auto& objective_fn = learner_parameters["objective"];
-    obj_->SaveConfig(&objective_fn);
-
-    std::vector<Json> metrics(metrics_.size());
-    for (size_t i = 0; i < metrics_.size(); ++i) {
-      metrics[i] = String(metrics_[i]->Name());
-    }
-    learner_parameters["metrics"] = Array(std::move(metrics));
-
-    learner_parameters["generic_param"] = toJson(generic_parameters_);
-  }
-
   // About to be deprecated by JSON format
   void LoadModel(dmlc::Stream* fi) override {
     generic_parameters_.UpdateAllowUnknown(Args{});
@@ -500,7 +663,7 @@ class LearnerImpl : public Learner {
 
     obj_.reset(ObjFunction::Create(tparam_.objective, &generic_parameters_));
     gbm_.reset(GradientBooster::Create(tparam_.booster, &generic_parameters_,
-                                       &learner_model_param_, cache_));
+                                       &learner_model_param_));
     gbm_->Load(fi);
     if (mparam_.contain_extra_attrs != 0) {
       std::vector<std::pair<std::string, std::string> > attr;
@@ -526,15 +689,23 @@ class LearnerImpl : public Learner {
       warn_old_model = false;
     }
 
-    if (mparam_.major_version >= 1) {
-      learner_model_param_ = LearnerModelParam(mparam_,
-                                               obj_->ProbToMargin(mparam_.base_score));
-    } else {
+    if (mparam_.major_version < 1) {
       // Before 1.0.0, base_score is saved as a transformed value, and there's no version
-      // attribute in the saved model.
-      learner_model_param_ = LearnerModelParam(mparam_, mparam_.base_score);
+      // attribute (saved a 0) in the saved model.
+      std::string multi{"multi:"};
+      if (!std::equal(multi.cbegin(), multi.cend(), tparam_.objective.cbegin())) {
+        HostDeviceVector<float> t;
+        t.HostVector().resize(1);
+        t.HostVector().at(0) = mparam_.base_score;
+        this->obj_->PredTransform(&t);
+        auto base_score = t.HostVector().at(0);
+        mparam_.base_score = base_score;
+      }
       warn_old_model = true;
     }
+
+    learner_model_param_ =
+        LearnerModelParam(mparam_, obj_->ProbToMargin(mparam_.base_score));
     if (attributes_.find("objective") != attributes_.cend()) {
       auto obj_str = attributes_.at("objective");
       auto j_obj = Json::Load({obj_str.c_str(), obj_str.size()});
@@ -572,7 +743,7 @@ class LearnerImpl : public Learner {
       tparam_.dsplit = DataSplitMode::kRow;
     }
 
-    this->Configure();
+    this->need_configuration_ = true;
   }
 
   // Save model into binary format.  The code is about to be deprecated by more robust
@@ -618,7 +789,8 @@ class LearnerImpl : public Learner {
       }
       extra_attr.emplace_back("metrics", os.str());
     }
-
+    std::string header {"binf"};
+    fo->Write(header.data(), 4);
     fo->Write(&mparam, sizeof(LearnerModelParamLegacy));
     fo->Write(tparam_.objective);
     fo->Write(tparam_.booster);
@@ -707,16 +879,41 @@ class LearnerImpl : public Learner {
       this->LoadConfig(config);
     }
   }
+};
+
+/*!
+ * \brief learner that performs gradient boosting for a specific objective
+ * function. It does training and prediction.
+ */
+class LearnerImpl : public LearnerIO {
+ public:
+  explicit LearnerImpl(std::vector<std::shared_ptr<DMatrix> > cache)
+      : LearnerIO{cache} {}
+  ~LearnerImpl() override {
+    auto local_map = XGBAPIThreadLocalStore::Get();
+    if (local_map->find(this) != local_map->cend()) {
+      local_map->erase(this);
+    }
+  }
+  // Configuration before data is known.
+  void CheckDataSplitMode() {
+    if (rabit::IsDistributed()) {
+      CHECK(tparam_.dsplit != DataSplitMode::kAuto)
+        << "Precondition violated; dsplit cannot be 'auto' in distributed mode";
+      if (tparam_.dsplit == DataSplitMode::kCol) {
+        LOG(FATAL) << "Column-wise data split is currently not supported.";
+      }
+    }
+  }
 
   std::vector<std::string> DumpModel(const FeatureMap& fmap,
                                      bool with_stats,
-                                     std::string format) const override {
-    CHECK(!this->need_configuration_)
-        << "The model hasn't been built yet.  Are you using raw Booster interface?";
+                                     std::string format) override {
+    this->Configure();
     return gbm_->DumpModel(fmap, with_stats, format);
   }
 
-  void UpdateOneIter(int iter, DMatrix* train) override {
+  void UpdateOneIter(int iter, std::shared_ptr<DMatrix> train) override {
     monitor_.Start("UpdateOneIter");
     TrainingObserver::Instance().Update(iter);
     this->Configure();
@@ -724,23 +921,25 @@ class LearnerImpl : public Learner {
       common::GlobalRandom().seed(generic_parameters_.seed * kRandSeedMagic + iter);
     }
     this->CheckDataSplitMode();
-    this->ValidateDMatrix(train);
+    this->ValidateDMatrix(train.get());
+
+    auto& predt = this->cache_.Cache(train, generic_parameters_.gpu_id);
 
     monitor_.Start("PredictRaw");
-    this->PredictRaw(train, &preds_[train], true);
+    this->PredictRaw(train.get(), &predt, true);
+    TrainingObserver::Instance().Observe(predt.predictions, "Predictions");
     monitor_.Stop("PredictRaw");
-    TrainingObserver::Instance().Observe(preds_[train], "Predictions");
 
     monitor_.Start("GetGradient");
-    obj_->GetGradient(preds_[train], train->Info(), iter, &gpair_);
+    obj_->GetGradient(predt.predictions, train->Info(), iter, &gpair_);
     monitor_.Stop("GetGradient");
     TrainingObserver::Instance().Observe(gpair_, "Gradients");
 
-    gbm_->DoBoost(train, &gpair_, obj_.get());
+    gbm_->DoBoost(train.get(), &gpair_, &predt);
     monitor_.Stop("UpdateOneIter");
   }
 
-  void BoostOneIter(int iter, DMatrix* train,
+  void BoostOneIter(int iter, std::shared_ptr<DMatrix> train,
                     HostDeviceVector<GradientPair>* in_gpair) override {
     monitor_.Start("BoostOneIter");
     this->Configure();
@@ -748,13 +947,15 @@ class LearnerImpl : public Learner {
       common::GlobalRandom().seed(generic_parameters_.seed * kRandSeedMagic + iter);
     }
     this->CheckDataSplitMode();
-    this->ValidateDMatrix(train);
+    this->ValidateDMatrix(train.get());
+    this->cache_.Cache(train, generic_parameters_.gpu_id);
 
-    gbm_->DoBoost(train, in_gpair);
+    gbm_->DoBoost(train.get(), in_gpair, &cache_.Entry(train.get()));
     monitor_.Stop("BoostOneIter");
   }
 
-  std::string EvalOneIter(int iter, const std::vector<DMatrix*>& data_sets,
+  std::string EvalOneIter(int iter,
+                          const std::vector<std::shared_ptr<DMatrix>>& data_sets,
                           const std::vector<std::string>& data_names) override {
     monitor_.Start("EvalOneIter");
     this->Configure();
@@ -766,14 +967,19 @@ class LearnerImpl : public Learner {
       metrics_.back()->Configure({cfg_.begin(), cfg_.end()});
     }
     for (size_t i = 0; i < data_sets.size(); ++i) {
-      DMatrix * dmat = data_sets[i];
-      this->ValidateDMatrix(dmat);
-      this->PredictRaw(data_sets[i], &preds_[dmat], false);
-      obj_->EvalTransform(&preds_[dmat]);
+      std::shared_ptr<DMatrix> m = data_sets[i];
+      auto &predt = this->cache_.Cache(m, generic_parameters_.gpu_id);
+      this->ValidateDMatrix(m.get());
+      this->PredictRaw(m.get(), &predt, false);
+
+      auto &out = output_predictions_.Cache(m, generic_parameters_.gpu_id).predictions;
+      out.Resize(predt.predictions.Size());
+      out.Copy(predt.predictions);
+
+      obj_->EvalTransform(&out);
       for (auto& ev : metrics_) {
         os << '\t' << data_names[i] << '-' << ev->Name() << ':'
-           << ev->Eval(preds_[dmat], data_sets[i]->Info(),
-                       tparam_.dsplit == DataSplitMode::kRow);
+           << ev->Eval(out, m->Info(), tparam_.dsplit == DataSplitMode::kRow);
       }
     }
 
@@ -781,56 +987,7 @@ class LearnerImpl : public Learner {
     return os.str();
   }
 
-  void SetParam(const std::string& key, const std::string& value) override {
-    this->need_configuration_ = true;
-    if (key == kEvalMetric) {
-      if (std::find(metric_names_.cbegin(), metric_names_.cend(),
-                    value) == metric_names_.cend()) {
-        metric_names_.emplace_back(value);
-      }
-    } else {
-      cfg_[key] = value;
-    }
-  }
-  // Short hand for setting multiple parameters
-  void SetParams(std::vector<std::pair<std::string, std::string>> const& args) override {
-    for (auto const& kv : args) {
-      this->SetParam(kv.first, kv.second);
-    }
-  }
-
-  void SetAttr(const std::string& key, const std::string& value) override {
-    attributes_[key] = value;
-    mparam_.contain_extra_attrs = 1;
-  }
-
-  bool GetAttr(const std::string& key, std::string* out) const override {
-    auto it = attributes_.find(key);
-    if (it == attributes_.end()) return false;
-    *out = it->second;
-    return true;
-  }
-
-  bool DelAttr(const std::string& key) override {
-    auto it = attributes_.find(key);
-    if (it == attributes_.end()) { return false; }
-    attributes_.erase(it);
-    return true;
-  }
-
-  std::vector<std::string> GetAttrNames() const override {
-    std::vector<std::string> out;
-    for (auto const& kv : attributes_) {
-      out.emplace_back(kv.first);
-    }
-    return out;
-  }
-
-  GenericParameter const& GetGenericParameter() const override {
-    return generic_parameters_;
-  }
-
-  void Predict(DMatrix* data, bool output_margin,
+  void Predict(std::shared_ptr<DMatrix> data, bool output_margin,
                HostDeviceVector<bst_float>* out_preds, unsigned ntree_limit,
                bool training,
                bool pred_leaf, bool pred_contribs, bool approx_contribs,
@@ -841,20 +998,45 @@ class LearnerImpl : public Learner {
     this->Configure();
     CHECK_LE(multiple_predictions, 1) << "Perform one kind of prediction at a time.";
     if (pred_contribs) {
-      gbm_->PredictContribution(data, &out_preds->HostVector(), ntree_limit, approx_contribs);
+      gbm_->PredictContribution(data.get(), &out_preds->HostVector(), ntree_limit, approx_contribs);
     } else if (pred_interactions) {
-      gbm_->PredictInteractionContributions(data, &out_preds->HostVector(), ntree_limit,
+      gbm_->PredictInteractionContributions(data.get(), &out_preds->HostVector(), ntree_limit,
                                             approx_contribs);
     } else if (pred_leaf) {
-      gbm_->PredictLeaf(data, &out_preds->HostVector(), ntree_limit);
+      gbm_->PredictLeaf(data.get(), &out_preds->HostVector(), ntree_limit);
     } else {
-      this->PredictRaw(data, out_preds, training, ntree_limit);
+      auto& prediction = cache_.Cache(data, generic_parameters_.gpu_id);
+      this->PredictRaw(data.get(), &prediction, training, ntree_limit);
+      // Copy the prediction cache to output prediction. out_preds comes from C API
+      out_preds->SetDevice(generic_parameters_.gpu_id);
+      out_preds->Resize(prediction.predictions.Size());
+      out_preds->Copy(prediction.predictions);
       if (!output_margin) {
         obj_->PredTransform(out_preds);
       }
     }
   }
 
+  XGBAPIThreadLocalEntry& GetThreadLocal() const override {
+    return (*XGBAPIThreadLocalStore::Get())[this];
+  }
+
+  void InplacePredict(dmlc::any const &x, std::string const &type,
+                      float missing, HostDeviceVector<bst_float> **out_preds,
+                      uint32_t layer_begin = 0, uint32_t layer_end = 0) override {
+    this->Configure();
+    auto& out_predictions = this->GetThreadLocal().prediction_entry;
+    this->gbm_->InplacePredict(x, missing, &out_predictions, layer_begin,
+                               layer_end);
+    if (type == "value") {
+      obj_->PredTransform(&out_predictions.predictions);
+    } else if (type == "margin") {
+    } else {
+      LOG(FATAL) << "Unsupported prediction type:" << type;
+    }
+    *out_preds = &out_predictions.predictions;
+  }
+
   const std::map<std::string, std::string>& GetConfigurationArguments() const override {
     return cfg_;
   }
@@ -868,153 +1050,38 @@ class LearnerImpl : public Learner {
    *   predictor, when it equals 0, this means we are using all the trees
    * \param training allow dropout when the DART booster is being used
    */
-  void PredictRaw(DMatrix* data, HostDeviceVector<bst_float>* out_preds,
+  void PredictRaw(DMatrix* data, PredictionCacheEntry* out_preds,
                   bool training,
                   unsigned ntree_limit = 0) const {
-    CHECK(gbm_ != nullptr)
-        << "Predict must happen after Load or configuration";
+    CHECK(gbm_ != nullptr) << "Predict must happen after Load or configuration";
     this->ValidateDMatrix(data);
     gbm_->PredictBatch(data, out_preds, training, ntree_limit);
   }
 
-  void ConfigureObjective(LearnerTrainParam const& old, Args* p_args) {
-    // Once binary IO is gone, NONE of these config is useful.
-    if (cfg_.find("num_class") != cfg_.cend() && cfg_.at("num_class") != "0" &&
-        tparam_.objective != "multi:softprob") {
-      cfg_["num_output_group"] = cfg_["num_class"];
-      if (atoi(cfg_["num_class"].c_str()) > 1 && cfg_.count("objective") == 0) {
-        tparam_.objective = "multi:softmax";
-      }
-    }
-
-    if (cfg_.find("max_delta_step") == cfg_.cend() &&
-        cfg_.find("objective") != cfg_.cend() &&
-        tparam_.objective == "count:poisson") {
-      // max_delta_step is a duplicated parameter in Poisson regression and tree param.
-      // Rename one of them once binary IO is gone.
-      cfg_["max_delta_step"] = kMaxDeltaStepDefaultValue;
-    }
-    if (obj_ == nullptr || tparam_.objective != old.objective) {
-      obj_.reset(ObjFunction::Create(tparam_.objective, &generic_parameters_));
-    }
-    auto& args = *p_args;
-    args = {cfg_.cbegin(), cfg_.cend()};  // renew
-    obj_->Configure(args);
-  }
-
-  void ConfigureMetrics(Args const& args) {
-    for (auto const& name : metric_names_) {
-      auto DupCheck = [&name](std::unique_ptr<Metric> const& m) {
-                        return m->Name() != name;
-                      };
-      if (std::all_of(metrics_.begin(), metrics_.end(), DupCheck)) {
-        metrics_.emplace_back(std::unique_ptr<Metric>(Metric::Create(name, &generic_parameters_)));
-        mparam_.contain_eval_metrics = 1;
-      }
-    }
-    for (auto& p_metric : metrics_) {
-      p_metric->Configure(args);
-    }
-  }
-
-  void ConfigureGBM(LearnerTrainParam const& old, Args const& args) {
-    if (gbm_ == nullptr || old.booster != tparam_.booster) {
-      gbm_.reset(GradientBooster::Create(tparam_.booster, &generic_parameters_,
-                                         &learner_model_param_, cache_));
-    }
-    gbm_->Configure(args);
-  }
-
-  // set number of features correctly.
-  void ConfigureNumFeatures() {
-    // estimate feature bound
-    // TODO(hcho3): Change num_feature to 64-bit integer
-    unsigned num_feature = 0;
-    for (auto & matrix : cache_) {
-      CHECK(matrix != nullptr);
-      const uint64_t num_col = matrix->Info().num_col_;
-      CHECK_LE(num_col, static_cast<uint64_t>(std::numeric_limits<unsigned>::max()))
-          << "Unfortunately, XGBoost does not support data matrices with "
-          << std::numeric_limits<unsigned>::max() << " features or greater";
-      num_feature = std::max(num_feature, static_cast<uint32_t>(num_col));
-    }
-    // run allreduce on num_feature to find the maximum value
-    rabit::Allreduce<rabit::op::Max>(&num_feature, 1, nullptr, nullptr, "num_feature");
-    if (num_feature > mparam_.num_feature) {
-      mparam_.num_feature = num_feature;
-    }
-    CHECK_NE(mparam_.num_feature, 0)
-        << "0 feature is supplied.  Are you using raw Booster interface?";
-    // Remove these once binary IO is gone.
-    cfg_["num_feature"] = common::ToString(mparam_.num_feature);
-    cfg_["num_class"] = common::ToString(mparam_.num_class);
-  }
-
   void ValidateDMatrix(DMatrix* p_fmat) const {
     MetaInfo const& info = p_fmat->Info();
-    auto const& weights = info.weights_;
-    if (info.group_ptr_.size() != 0 && weights.Size() != 0) {
-      CHECK(weights.Size() == info.group_ptr_.size() - 1)
-          << "\n"
-          << "weights size: " << weights.Size()            << ", "
-          << "groups size: "  << info.group_ptr_.size() -1 << ", "
-          << "num rows: "     << p_fmat->Info().num_row_   << "\n"
-          << "Number of weights should be equal to number of groups in ranking task.";
-    }
+    info.Validate(generic_parameters_.gpu_id);
 
     auto const row_based_split = [this]() {
       return tparam_.dsplit == DataSplitMode::kRow ||
              tparam_.dsplit == DataSplitMode::kAuto;
     };
-    bool const valid_features =
-        !row_based_split() ||
-        (learner_model_param_.num_feature == p_fmat->Info().num_col_);
-    std::string const msg {
-      "Number of columns does not match number of features in booster."
-    };
-    if (generic_parameters_.validate_features) {
-      CHECK_EQ(learner_model_param_.num_feature, p_fmat->Info().num_col_) << msg;
-    } else if (!valid_features) {
-      // Remove this and make the equality check fatal once spark can fix all failing tests.
-      LOG(WARNING) << msg << " "
-                   << "Columns: " << p_fmat->Info().num_col_ << " "
-                   << "Features: " << learner_model_param_.num_feature;
+    if (row_based_split()) {
+      CHECK_EQ(learner_model_param_.num_feature, p_fmat->Info().num_col_)
+          << "Number of columns does not match number of features in booster.";
     }
   }
 
-  // model parameter
-  LearnerModelParamLegacy mparam_;
-  LearnerModelParam learner_model_param_;
-  LearnerTrainParam tparam_;
-  // Used to identify the offset of JSON string when
-  // `enable_experimental_json_serialization' is set to false.  Will be removed once JSON
-  // takes over.
-  std::string const serialisation_header_ { u8"CONFIG-offset:" };
-  // configurations
-  std::map<std::string, std::string> cfg_;
-  std::map<std::string, std::string> attributes_;
-  std::vector<std::string> metric_names_;
-  static std::string const kEvalMetric;  // NOLINT
-  // temporal storages for prediction
-  std::map<DMatrix*, HostDeviceVector<bst_float>> preds_;
-  // gradient pairs
-  HostDeviceVector<GradientPair> gpair_;
-  bool need_configuration_;
-
  private:
   /*! \brief random number transformation seed. */
   static int32_t constexpr kRandSeedMagic = 127;
-  // internal cached dmatrix
-  std::vector<std::shared_ptr<DMatrix> > cache_;
-
-  common::Monitor monitor_;
-
-  /*! \brief (Deprecated) saved config keys used to restore failed worker */
-  std::set<std::string> saved_configs_ = {"num_round"};
+  // gradient pairs
+  HostDeviceVector<GradientPair> gpair_;
+  /*! \brief Temporary storage to prediction.  Useful for storing data transformed by
+   *  objective function */
+  PredictionContainer output_predictions_;
 };
 
-std::string const LearnerImpl::kEvalMetric {"eval_metric"};  // NOLINT
-
 constexpr int32_t LearnerImpl::kRandSeedMagic;
 
 Learner* Learner::Create(
